@@ -1,10 +1,11 @@
-r"""Parameterized flows and autoregressive transformations."""
+r"""Parameterized flows and transformations."""
 
 __all__ = [
     'DistributionModule',
     'TransformModule',
     'FlowModule',
     'GMM',
+    'ElementWiseTransform',
     'MaskedAutoregressiveTransform',
     'MAF',
     'NSF',
@@ -14,6 +15,8 @@ __all__ = [
     'NAF',
     'UnconstrainedNeuralAutoregressiveTransform',
     'UNAF',
+    'GeneralCouplingTransform',
+    'GCF',
     'FFJTransform',
     'CNF',
 ]
@@ -23,29 +26,29 @@ import torch
 import torch.nn as nn
 
 from functools import partial
-from math import ceil, pi
+from math import ceil, pi, prod
 from textwrap import indent
-from torch import Tensor, LongTensor, Size
+from torch import Tensor, BoolTensor, LongTensor, Size
 from torch.distributions import *
 from typing import *
 
 from .distributions import *
 from .transforms import *
 from .nn import *
-from .utils import broadcast
+from .utils import broadcast, unpack
 
 
 class DistributionModule(nn.Module, abc.ABC):
     r"""Abstract distribution module."""
 
     @abc.abstractmethod
-    def forward(y: Tensor = None) -> Distribution:
+    def forward(c: Tensor = None) -> Distribution:
         r"""
         Arguments:
-            y: A context :math:`y`.
+            c: A context :math:`c`.
 
         Returns:
-            A distribution :math:`p(X | y)`.
+            A distribution :math:`p(X | c)`.
         """
 
         pass
@@ -55,13 +58,13 @@ class TransformModule(nn.Module, abc.ABC):
     r"""Abstract transformation module."""
 
     @abc.abstractmethod
-    def forward(y: Tensor = None) -> Transform:
+    def forward(c: Tensor = None) -> Transform:
         r"""
         Arguments:
-            y: A context :math:`y`.
+            c: A context :math:`c`.
 
         Returns:
-            A transformation :math:`z = f(x | y)`.
+            A transformation :math:`y = f(x | c)`.
         """
 
         pass
@@ -85,21 +88,21 @@ class FlowModule(DistributionModule):
         self.transforms = nn.ModuleList(transforms)
         self.base = base
 
-    def forward(self, y: Tensor = None) -> NormalizingFlow:
+    def forward(self, c: Tensor = None) -> NormalizingFlow:
         r"""
         Arguments:
-            y: A context :math:`y`.
+            c: A context :math:`c`.
 
         Returns:
-            A normalizing flow :math:`p(X | y)`.
+            A normalizing flow :math:`p(X | c)`.
         """
 
-        transform = ComposedTransform(*(t(y) for t in self.transforms))
+        transform = ComposedTransform(*(t(c) for t in self.transforms))
 
-        if y is None:
-            base = self.base(y)
+        if c is None:
+            base = self.base(c)
         else:
-            base = self.base(y).expand(y.shape[:-1])
+            base = self.base(c).expand(c.shape[:-1])
 
         return NormalizingFlow(transform, base)
 
@@ -137,7 +140,7 @@ class Unconditional(nn.Module):
     def __repr__(self) -> str:
         return repr(self.forward())
 
-    def forward(self, y: Tensor = None) -> Any:
+    def forward(self, c: Tensor = None) -> Any:
         return self.meta(
             *self._parameters.values(),
             *self._buffers.values(),
@@ -150,7 +153,7 @@ class Parameters(nn.ParameterList):
 
     def extra_repr(self) -> str:
         lines = [
-            f'({i}): tensor of shape {tuple(p.shape)}'
+            f'({i}): Tensor(shape={tuple(p.shape)})'
             for i, p in enumerate(self)
         ]
 
@@ -160,13 +163,13 @@ class Parameters(nn.ParameterList):
 class GMM(DistributionModule):
     r"""Creates a Gaussian mixture model (GMM).
 
-    .. math:: p(X | y) = \sum_{i = 1}^K w_i(y) \, \mathcal{N}(X | \mu_i(y), \Sigma_i(y))
+    .. math:: p(X | c) = \sum_{i = 1}^K w_i(c) \, \mathcal{N}(X | \mu_i(c), \Sigma_i(c))
 
     Arguments:
         features: The number of features.
         context: The number of context features.
         components: The number of components :math:`K` in the mixture.
-        **kwargs: Keyword arguments passed to :class:`zuko.nn.MLP`.
+        kwargs: Keyword arguments passed to :class:`zuko.nn.MLP`.
     """
 
     def __init__(
@@ -185,21 +188,20 @@ class GMM(DistributionModule):
             (components, features * (features - 1) // 2),  # off diagonal
         ]
 
-        self.shapes = list(map(Size, shapes))
-        self.sizes = [s.numel() for s in self.shapes]
+        self.shapes = shapes
+        self.total = sum(prod(s) for s in shapes)
 
         if context > 0:
-            self.hyper = MLP(context, sum(self.sizes), **kwargs)
+            self.hyper = MLP(context, self.total, **kwargs)
         else:
             self.phi = Parameters(torch.randn(*s) for s in shapes)
 
-    def forward(self, y: Tensor = None) -> Distribution:
-        if y is None:
+    def forward(self, c: Tensor = None) -> Distribution:
+        if c is None:
             phi = self.phi
         else:
-            phi = self.hyper(y)
-            phi = phi.split(self.sizes, -1)
-            phi = (p.unflatten(-1, s) for p, s in zip(phi, self.shapes))
+            phi = self.hyper(c)
+            phi = unpack(phi, self.shapes)
 
         logits, loc, diag, tril = phi
 
@@ -210,6 +212,75 @@ class GMM(DistributionModule):
         return Mixture(MultivariateNormal(loc=loc, scale_tril=scale), logits)
 
 
+class ElementWiseTransform(TransformModule):
+    r"""Creates an element-wise transformation.
+
+    Arguments:
+        features: The number of features.
+        context: The number of context features.
+        univariate: The univariate transformation constructor.
+        shapes: The shapes of the univariate transformation parameters.
+        kwargs: Keyword arguments passed to :class:`zuko.nn.MLP`.
+
+    Example:
+        >>> t = ElementWiseTransform(3, 4)
+        >>> t
+        ElementWiseTransform(
+          (base): MonotonicAffineTransform()
+          (hyper): MLP(
+            (0): Linear(in_features=4, out_features=64, bias=True)
+            (1): ReLU()
+            (2): Linear(in_features=64, out_features=64, bias=True)
+            (3): ReLU()
+            (4): Linear(in_features=64, out_features=6, bias=True)
+          )
+        )
+        >>> x = torch.randn(3)
+        >>> x
+        tensor([2.1983,  -1.3182,  0.0329])
+        >>> c = torch.randn(4)
+        >>> y = t(c)(x)
+        >>> t(c).inv(y)
+        tensor([2.1983,  -1.3182,  0.0329])
+    """
+
+    def __init__(
+        self,
+        features: int,
+        context: int = 0,
+        univariate: Callable[..., Transform] = MonotonicAffineTransform,
+        shapes: Sequence[Size] = ((), ()),
+        **kwargs,
+    ):
+        super().__init__()
+
+        self.univariate = univariate
+        self.shapes = shapes
+        self.total = sum(prod(s) for s in shapes)
+
+        if context > 0:
+            self.hyper = MLP(context, features * self.total, **kwargs)
+        else:
+            self.phi = Parameters(torch.randn(features, *s) for s in shapes)
+
+    def extra_repr(self) -> str:
+        base = self.univariate(*map(torch.randn, self.shapes))
+
+        return '\n'.join([
+            f'(base): {base}',
+        ])
+
+    def forward(self, c: Tensor = None) -> Transform:
+        if c is None:
+            phi = self.phi
+        else:
+            phi = self.hyper(c)
+            phi = phi.unflatten(-1, (-1, self.total))
+            phi = unpack(phi, self.shapes)
+
+        return DependentTransform(self.univariate(*phi), 1)
+
+
 class MaskedAutoregressiveTransform(TransformModule):
     r"""Creates a masked autoregressive transformation.
 
@@ -245,9 +316,9 @@ class MaskedAutoregressiveTransform(TransformModule):
         >>> x = torch.randn(3)
         >>> x
         tensor([-0.9485,  1.5290,  0.2018])
-        >>> y = torch.randn(4)
-        >>> z = t(y)(x)
-        >>> t(y).inv(z)
+        >>> c = torch.randn(4)
+        >>> y = t(c)(x)
+        >>> t(c).inv(y)
         tensor([-0.9485,  1.5290,  0.2018])
     """
 
@@ -265,8 +336,8 @@ class MaskedAutoregressiveTransform(TransformModule):
 
         # Univariate transformation
         self.univariate = univariate
-        self.shapes = list(map(Size, shapes))
-        self.sizes = [s.numel() for s in self.shapes]
+        self.shapes = shapes
+        self.total = sum(prod(s) for s in shapes)
 
         # Adjacency
         self.register_buffer('order', None)
@@ -283,7 +354,7 @@ class MaskedAutoregressiveTransform(TransformModule):
         self.order = torch.div(order, ceil(features / self.passes), rounding_mode='floor')
 
         in_order = torch.cat((self.order, torch.full((context,), -1)))
-        out_order = torch.repeat_interleave(self.order, sum(self.sizes))
+        out_order = torch.repeat_interleave(self.order, self.total)
         adjacency = out_order[:, None] > in_order
 
         # Hyper network
@@ -294,29 +365,26 @@ class MaskedAutoregressiveTransform(TransformModule):
         order = self.order.tolist()
 
         if len(order) > 10:
-            order = str(order[:5] + [...] + order[-5:]).replace('Ellipsis', '...')
+            order = order[:5] + [...] + order[-5:]
+            order = str(order).replace('Ellipsis', '...')
 
         return '\n'.join([
             f'(base): {base}',
             f'(order): {order}',
         ])
 
-    def meta(self, y: Tensor, x: Tensor) -> Transform:
-        if y is not None:
-            x = torch.cat(broadcast(x, y, ignore=1), dim=-1)
-
-        total = sum(self.sizes)
+    def meta(self, c: Tensor, x: Tensor) -> Transform:
+        if c is not None:
+            x = torch.cat(broadcast(x, c, ignore=1), dim=-1)
 
         phi = self.hyper(x)
-        phi = phi.unflatten(-1, (phi.shape[-1] // total, total))
-        phi = phi.split(self.sizes, -1)
-        phi = (p.unflatten(-1, s + (1,)) for p, s in zip(phi, self.shapes))
-        phi = (p.squeeze(-1) for p in phi)
+        phi = phi.unflatten(-1, (-1, self.total))
+        phi = unpack(phi, self.shapes)
 
-        return self.univariate(*phi)
+        return DependentTransform(self.univariate(*phi), 1)
 
-    def forward(self, y: Tensor = None) -> Transform:
-        return AutoregressiveTransform(partial(self.meta, y), self.passes)
+    def forward(self, c: Tensor = None) -> Transform:
+        return AutoregressiveTransform(partial(self.meta, c), self.passes)
 
 
 class MAF(FlowModule):
@@ -376,11 +444,11 @@ class MAF(FlowModule):
           )
           (base): DiagNormal(loc: torch.Size([3]), scale: torch.Size([3]))
         )
-        >>> y = torch.randn(4)
-        >>> x = flow(y).sample()
+        >>> c = torch.randn(4)
+        >>> x = flow(c).sample()
         >>> x
         tensor([-1.7154, -0.4401,  0.7505])
-        >>> flow(y).log_prob(x)
+        >>> flow(c).log_prob(x)
         tensor(-4.4630, grad_fn=<AddBackward0>)
     """
 
@@ -526,7 +594,7 @@ class SOSPF(MAF):
         )
 
         for i in reversed(range(len(self.transforms))):
-            self.transforms.insert(i, Unconditional(SoftclipTransform))
+            self.transforms.insert(i, Unconditional(SoftclipTransform, bound=6.0))
 
 
 class NeuralAutoregressiveTransform(MaskedAutoregressiveTransform):
@@ -572,9 +640,9 @@ class NeuralAutoregressiveTransform(MaskedAutoregressiveTransform):
         >>> x = torch.randn(3)
         >>> x
         tensor([-2.3267,  1.4581, -1.6776])
-        >>> y = torch.randn(4)
-        >>> z = t(y)(x)
-        >>> t(y).inv(z)
+        >>> c = torch.randn(4)
+        >>> y = t(c)(x)
+        >>> t(c).inv(y)
         tensor([-2.3267,  1.4581, -1.6776])
     """
 
@@ -650,7 +718,7 @@ class NAF(FlowModule):
         ]
 
         for i in reversed(range(len(transforms))):
-            transforms.insert(i, Unconditional(SoftclipTransform))
+            transforms.insert(i, Unconditional(SoftclipTransform, bound=6.0))
 
         base = Unconditional(
             DiagNormal,
@@ -706,9 +774,9 @@ class UnconstrainedNeuralAutoregressiveTransform(MaskedAutoregressiveTransform):
         >>> x = torch.randn(3)
         >>> x
         tensor([-0.0103, -1.0871, -0.0667])
-        >>> y = torch.randn(4)
-        >>> z = t(y)(x)
-        >>> t(y).inv(z)
+        >>> c = torch.randn(4)
+        >>> y = t(c)(x)
+        >>> t(c).inv(y)
         tensor([-0.0103, -1.0871, -0.0667])
     """
 
@@ -787,7 +855,7 @@ class UNAF(FlowModule):
         ]
 
         for i in reversed(range(len(transforms))):
-            transforms.insert(i, Unconditional(SoftclipTransform))
+            transforms.insert(i, Unconditional(SoftclipTransform, bound=6.0))
 
         base = Unconditional(
             DiagNormal,
@@ -799,6 +867,150 @@ class UNAF(FlowModule):
         super().__init__(transforms, base)
 
 
+class GeneralCouplingTransform(TransformModule):
+    r"""Creates a general coupling transformation.
+
+    References:
+        | NICE: Non-linear Independent Components Estimation (Dinh et al., 2014)
+        | https://arxiv.org/abs/1410.8516
+
+    Arguments:
+        features: The number of features.
+        context: The number of context features.
+        mask: The coupling mask. If :py:`None`, use a checkered mask.
+        univariate: The univariate transformation constructor.
+        shapes: The shapes of the univariate transformation parameters.
+        kwargs: Keyword arguments passed to :class:`zuko.nn.MLP`.
+
+    Example:
+        >>> t = GeneralCouplingTransform(3, 4)
+        >>> t
+        GeneralCouplingTransform(
+          (base): MonotonicAffineTransform()
+          (order): [0, 1, 0]
+          (hyper): MLP(
+            (0): Linear(in_features=5, out_features=64, bias=True)
+            (1): ReLU()
+            (2): Linear(in_features=64, out_features=64, bias=True)
+            (3): ReLU()
+            (4): Linear(in_features=64, out_features=4, bias=True)
+          )
+        )
+        >>> x = torch.randn(3)
+        >>> x
+        tensor([-0.8743,  0.6232,  1.2439])
+        >>> c = torch.randn(4)
+        >>> y = t(c)(x)
+        >>> t(c).inv(y)
+        tensor([-0.8743,  0.6232,  1.2439])
+    """
+
+    def __init__(
+        self,
+        features: int,
+        context: int = 0,
+        mask: BoolTensor = None,
+        univariate: Callable[..., Transform] = MonotonicAffineTransform,
+        shapes: Sequence[Size] = ((), ()),
+        **kwargs,
+    ):
+        super().__init__()
+
+        # Univariate transformation
+        self.univariate = univariate
+        self.shapes = shapes
+        self.total = sum(prod(s) for s in shapes)
+
+        # Mask
+        self.register_buffer('mask', None)
+
+        if mask is None:
+            self.mask = torch.arange(features) % 2 == 1
+        else:
+            self.mask = mask
+
+        features_a = self.mask.sum().item()
+        features_b = features - features_a
+
+        # Hyper network
+        self.hyper = MLP(features_a + context, features_b * self.total, **kwargs)
+
+    def extra_repr(self) -> str:
+        base = self.univariate(*map(torch.randn, self.shapes))
+        mask = self.mask.int().tolist()
+
+        if len(mask) > 10:
+            mask = mask[:5] + [...] + mask[-5:]
+            mask = str(mask).replace('Ellipsis', '...')
+
+        return '\n'.join([
+            f'(base): {base}',
+            f'(mask): {mask}',
+        ])
+
+    def meta(self, c: Tensor, x: Tensor) -> Transform:
+        if c is not None:
+            x = torch.cat(broadcast(x, c, ignore=1), dim=-1)
+
+        phi = self.hyper(x)
+        phi = phi.unflatten(-1, (-1, self.total))
+        phi = unpack(phi, self.shapes)
+
+        return DependentTransform(self.univariate(*phi), 1)
+
+    def forward(self, c: Tensor = None) -> Transform:
+        return CouplingTransform(partial(self.meta, c), self.mask)
+
+
+class GCF(FlowModule):
+    r"""Creates a general coupling flow (GCF).
+
+    Arguments:
+        features: The number of features.
+        context: The number of context features.
+        transforms: The number of coupling transformations.
+        randmask: Whether random coupling masks are used or not. If :py:`False`,
+            use alternating checkered masks.
+        features are in ascending (descending) order for even
+            (odd) transformations.
+        kwargs: Keyword arguments passed to :class:`GeneralCouplingTransform`.
+    """
+
+    def __init__(
+        self,
+        features: int,
+        context: int = 0,
+        transforms: int = 3,
+        randmask: bool = False,
+        **kwargs,
+    ):
+        temp = []
+
+        for i in range(transforms):
+            if randmask:
+                mask = torch.randperm(features) % 2 == i % 2
+            else:
+                mask = torch.arange(features) % 2 == i % 2
+
+            temp.append(
+                GeneralCouplingTransform(
+                    features=features,
+                    context=context,
+                    mask=mask,
+                    **kwargs,
+                )
+            )
+
+        base = Unconditional(
+            DiagNormal,
+            torch.zeros(features),
+            torch.ones(features),
+            buffer=True,
+        )
+
+        super().__init__(temp, base)
+
+
 class FFJTransform(TransformModule):
     r"""Creates a free-form Jacobian (FFJ) transformation.
 
@@ -809,7 +1021,7 @@ class FFJTransform(TransformModule):
     Arguments:
         features: The number of features.
         context: The number of context features.
-        frequencies: The number of time embedding frequencies.
+        freqs: The number of time embedding frequencies.
         exact: Whether the exact log-determinant of the Jacobian or an unbiased
             stochastic estimate thereof is calculated.
         kwargs: Keyword arguments passed to :class:`zuko.nn.MLP`.
@@ -829,9 +1041,9 @@ class FFJTransform(TransformModule):
         >>> x = torch.randn(3)
         >>> x
         tensor([ 0.1777,  1.0139, -1.0370])
-        >>> y = torch.randn(4)
-        >>> z = t(y)(x)
-        >>> t(y).inv(z)
+        >>> c = torch.randn(4)
+        >>> y = t(c)(x)
+        >>> t(c).inv(y)
         tensor([ 0.1777,  1.0139, -1.0370])
     """
 
@@ -839,7 +1051,7 @@ class FFJTransform(TransformModule):
         self,
         features: int,
         context: int = 0,
-        frequencies: int = 3,
+        freqs: int = 3,
         exact: bool = True,
         **kwargs,
     ):
@@ -847,29 +1059,30 @@ class FFJTransform(TransformModule):
 
         kwargs.setdefault('activation', nn.ELU)
 
-        self.ode = MLP(features + context + 2 * frequencies, features, **kwargs)
+        self.ode = MLP(features + context + 2 * freqs, features, **kwargs)
 
-        self.register_buffer('time', torch.tensor(1.0))
-        self.register_buffer('frequencies', 2 ** torch.arange(frequencies) * pi)
+        self.register_buffer('times', torch.tensor((0.0, 1.0)))
+        self.register_buffer('freqs', torch.arange(1, freqs + 1) * pi)
 
         self.exact = exact
 
-    def f(self, t: Tensor, x: Tensor, y: Tensor = None) -> Tensor:
-        t = self.frequencies * t[..., None]
+    def f(self, t: Tensor, x: Tensor, c: Tensor = None) -> Tensor:
+        t = self.freqs * t[..., None]
         t = torch.cat((t.cos(), t.sin()), dim=-1)
 
-        if y is None:
+        if c is None:
             x = torch.cat(broadcast(t, x, ignore=1), dim=-1)
         else:
-            x = torch.cat(broadcast(t, x, y, ignore=1), dim=-1)
+            x = torch.cat(broadcast(t, x, c, ignore=1), dim=-1)
 
         return self.ode(x)
 
-    def forward(self, y: Tensor = None) -> Transform:
+    def forward(self, c: Tensor = None) -> Transform:
         return FreeFormJacobianTransform(
-            f=partial(self.f, y=y),
-            time=self.time,
-            phi=self.ode.parameters() if y is None else (y, *self.ode.parameters()),
+            f=partial(self.f, c=c),
+            t0=self.times[0],
+            t1=self.times[1],
+            phi=self.parameters() if c is None else (c, *self.parameters()),
             exact=self.exact,
         )
 
@@ -888,7 +1101,6 @@ class CNF(FlowModule):
     Arguments:
         features: The number of features.
         context: The number of context features.
-        transforms: The number of transformations.
         kwargs: Keyword arguments passed to :class:`FFJTransform`.
     """
 
@@ -896,7 +1108,6 @@ class CNF(FlowModule):
         self,
         features: int,
         context: int = 0,
-        transforms: int = 1,
         **kwargs,
     ):
         transforms = [
@@ -905,7 +1116,6 @@ class CNF(FlowModule):
                 context=context,
                 **kwargs,
             )
-            for _ in range(transforms)
         ]
 
         base = Unconditional(
