@@ -2,6 +2,7 @@ r"""Parameterizable transformations."""
 
 __all__ = [
     'ComposedTransform',
+    'DependentTransform',
     'IdentityTransform',
     'CosTransform',
     'SinTransform',
@@ -14,6 +15,7 @@ __all__ = [
     'SOSPolynomialTransform',
     'FreeFormJacobianTransform',
     'AutoregressiveTransform',
+    'CouplingTransform',
     'LULinearTransform',
     'PermutationTransform',
 ]
@@ -23,7 +25,7 @@ import torch
 import torch.nn.functional as F
 
 from textwrap import indent
-from torch import Tensor, LongTensor, Size
+from torch import Tensor, BoolTensor, LongTensor, Size
 from torch.distributions import *
 from torch.distributions import constraints
 from torch.distributions.utils import _sum_rightmost
@@ -51,6 +53,8 @@ Transform.call_and_ladj = _call_and_ladj
 class ComposedTransform(Transform):
     r"""Creates a transformation :math:`f(x) = f_n \circ \dots \circ f_0(x)`.
 
+    Optimized version of :class:`torch.distributions.transforms.ComposeTransform`.
+
     Arguments:
         transforms: A sequence of transformations :math:`f_i`.
     """
@@ -150,6 +154,66 @@ class ComposedTransform(Transform):
         return shape
 
 
+class DependentTransform(Transform):
+    r"""Wraps a base transformation to treat right-most dimensions as dependent.
+
+    Optimized version of :class:`torch.distributions.transforms.IndependentTransform`.
+
+    Arguments:
+        base: The base transformation.
+        reinterpreted: The number of dimensions to treat as dependent.
+    """
+
+    def __init__(self, base: Transform, reinterpreted: int, **kwargs):
+        super().__init__(**kwargs)
+
+        self.base = base
+        self.reinterpreted = reinterpreted
+
+    def __repr__(self) -> str:
+        return f'{self.__class__.__name__}({self.base}, {self.reinterpreted})'
+
+    @property
+    def domain(self) -> constraints.Constraint:
+        return constraints.independent(self.base.domain, self.reinterpreted)
+
+    @property
+    def codomain(self) -> constraints.Constraint:
+        return constraints.independent(self.base.codomain, self.reinterpreted)
+
+    @property
+    def bijective(self) -> bool:
+        return self.base.bijective
+
+    def _call(self, x: Tensor) -> Tensor:
+        return self.base(x)
+
+    @property
+    def inv(self) -> Transform:
+        return DependentTransform(self.base.inv, self.reinterpreted)
+
+    def _inverse(self, y: Tensor) -> Tensor:
+        return self.base.inv(y)
+
+    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
+        ladj = self.base.log_abs_det_jacobian(x, y)
+        ladj = _sum_rightmost(ladj, self.reinterpreted)
+
+        return ladj
+
+    def call_and_ladj(self, x: Tensor) -> Tuple[Tensor, Tensor]:
+        y, ladj = self.base.call_and_ladj(x)
+        ladj = ladj = _sum_rightmost(ladj, self.reinterpreted)
+
+        return y, ladj
+
+    def forward_shape(self, shape: Size) -> Size:
+        return self.base.forward_shape(shape)
+
+    def inverse_shape(self, shape: Size) -> Size:
+        return self.base.inverse_shape(shape)
+
+
 class IdentityTransform(Transform):
     r"""Creates a transformation :math:`f(x) = x`."""
 
@@ -597,7 +661,7 @@ class FreeFormJacobianTransform(Transform):
     The transformation is the integration of a system of first-order ordinary
     differential equations
 
-    .. math:: x(T) = x(0) + \int_0^T f_\phi(t, x(t)) ~ dt .
+    .. math:: x(t_1) = x_0 + \int_{t_0}^{t_1} f_\phi(t, x(t)) ~ dt .
 
     References:
         | FFJORD: Free-form Continuous Dynamics for Scalable Reversible Generative Models (Grathwohl et al., 2018)
@@ -605,7 +669,8 @@ class FreeFormJacobianTransform(Transform):
 
     Arguments:
         f: A system of first-order ODEs :math:`f_\phi`.
-        time: The integration time :math:`T`.
+        t0: The initial integration time :math:`t_0`.
+        t1: The final integration time :math:`t_1`.
         phi: The parameters :math:`\phi` of :math:`f_\phi`.
         exact: Whether the exact log-determinant of the Jacobian or an unbiased
             stochastic estimate thereof is calculated.
@@ -618,7 +683,8 @@ class FreeFormJacobianTransform(Transform):
     def __init__(
         self,
         f: Callable[[Tensor, Tensor], Tensor],
-        time: Tensor,
+        t0: Union[float, Tensor] = 0.0,
+        t1: Union[float, Tensor] = 1.0,
         phi: Iterable[Tensor] = (),
         exact: bool = True,
         **kwargs,
@@ -626,8 +692,8 @@ class FreeFormJacobianTransform(Transform):
         super().__init__(**kwargs)
 
         self.f = f
-        self.t0 = time.new_tensor(0.0)
-        self.t1 = time
+        self.t0 = t0
+        self.t1 = t1
         self.phi = tuple(filter(lambda p: p.requires_grad, phi))
         self.exact = exact
         self.trace_scale = 1e-2  # relax jacobian tolerances
@@ -637,17 +703,13 @@ class FreeFormJacobianTransform(Transform):
 
     @property
     def inv(self) -> Transform:
-        new = self.__new__(FreeFormJacobianTransform)
-        new.f = self.f
-        new.t0 = self.t1
-        new.t1 = self.t0
-        new.phi = self.phi
-        new.exact = self.exact
-        new.trace_scale = self.trace_scale
-
-        Transform.__init__(new)
-
-        return new
+        return FreeFormJacobianTransform(
+            f=self.f,
+            t0=self.t1,
+            t1=self.t0,
+            phi=self.phi,
+            exact=self.exact,
+        )
 
     def _inverse(self, y: Tensor) -> Tensor:
         return odeint(self.f, y, self.t1, self.t0, self.phi)
@@ -719,11 +781,75 @@ class AutoregressiveTransform(Transform):
         return x
 
     def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
-        return self.meta(x).log_abs_det_jacobian(x, y).sum(dim=-1)
+        return self.meta(x).log_abs_det_jacobian(x, y)
 
     def call_and_ladj(self, x: Tensor) -> Tuple[Tensor, Tensor]:
         y, ladj = self.meta(x).call_and_ladj(x)
-        return y, ladj.sum(dim=-1)
+        return y, ladj
+
+
+class CouplingTransform(Transform):
+    r"""Transform via a coupling scheme.
+
+    .. math::
+        y_a & = x_a \\
+        y_b & = f(x_b; x_a)
+
+    Arguments:
+        meta: A meta function which returns a transformation :math:`f`.
+        mask: A coupling mask defining the split $x \mapsto (x_a, x_b)$.
+    """
+
+    domain = constraints.real_vector
+    codomain = constraints.real_vector
+    bijective = True
+
+    def __init__(
+        self,
+        meta: Callable[[Tensor], Transform],
+        mask: BoolTensor,
+        **kwargs,
+    ):
+        super().__init__(**kwargs)
+
+        self.meta = meta
+        self.idx_a = mask.nonzero().squeeze(-1)
+        self.idx_b = (~mask).nonzero().squeeze(-1)
+
+    def split(self, x: Tensor) -> Tuple[Tensor, Tensor]:
+        return x[..., self.idx_a], x[..., self.idx_b]
+
+    def merge(self, x_a: Tensor, x_b: Tensor, shape: Size) -> Tensor:
+        x = x_a.new_empty(shape)
+        x[..., self.idx_a] = x_a
+        x[..., self.idx_b] = x_b
+
+        return x
+
+    def _call(self, x: Tensor) -> Tensor:
+        x_a, x_b = self.split(x)
+        y_b = self.meta(x_a)(x_b)
+
+        return self.merge(x_a, y_b, x.shape)
+
+    def _inverse(self, y: Tensor) -> Tensor:
+        y_a, y_b = self.split(y)
+        x_b = self.meta(y_a).inv(y_b)
+
+        return self.merge(y_a, x_b, y.shape)
+
+    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
+        x_a, x_b = self.split(x)
+        _, y_b = self.split(y)
+
+        return self.meta(x_a).log_abs_det_jacobian(x_b, y_b)
+
+    def call_and_ladj(self, x: Tensor) -> Tuple[Tensor, Tensor]:
+        x_a, x_b = self.split(x)
+        y_b, ladj = self.meta(x_a).call_and_ladj(x_b)
+        y = self.merge(x_a, y_b, x.shape)
+
+        return y, ladj
 
 
 class LULinearTransform(Transform):
@@ -786,7 +912,8 @@ class PermutationTransform(Transform):
         order = self.order.tolist()
 
         if len(order) > 10:
-            order = str(order[:5] + [...] + order[-5:]).replace('Ellipsis', '...')
+            order = order[:5] + [...] + order[-5:]
+            order = str(order).replace('Ellipsis', '...')
 
         return f'{self.__class__.__name__}({order})'
 
